@import url("resource:///org/gnome/builder/themes/shared/shared-buildui.css");
@import url("resource:///org/gnome/builder/themes/shared/shared-completion.css");
@import url("resource:///org/gnome/builder/themes/shared/shared-debugger.css");
@import url("resource:///org/gnome/builder/themes/shared/shared-layout.css");
@import url("resource:///org/gnome/builder/themes/shared/shared-editor.css");
@import url("resource:///org/gnome/builder/themes/shared/shared-greeter.css");
@import url("resource:///org/gnome/builder/themes/shared/shared-hoverer.css");
@import url("resource:///org/gnome/builder/themes/shared/shared-omnibar.css");
@import url("resource:///org/gnome/builder/themes/shared/shared-search.css");
@import url("resource:///org/gnome/builder/themes/shared/shared-treeview.css");

/* work around some gtk padding issue */
filechooser actionbar button.combo {
  padding: 0;
}

/* Generic styles */
.warning {
    color: @warning_color;
}

/* headerbar shadow in fullscreen, for depth above children */
window.fullscreen headerbar {
  margin-bottom: 5px;
  box-shadow: 0 0 3px 3px alpha(@wm_shadow,.3);
}


/* Styling in the create-project surface. */
createprojectsurface stack > box:first-child list row {
  padding: 10px;
  border-bottom: 1px solid alpha(@borders, 0.2);
}
createprojectsurface stack > box:first-child list row:last-child {
  border-bottom: none;
}

/*
 * Perspectives switcher
 *
 * The following tweaks the left-most sidebar containing
 * the list of surfaces.
 */
surfaceswitcher {
  border-right: 1px solid alpha(@borders, 0.5);
}
surfaceswitcher button {
  background: transparent;
  border-radius: 0;
  border: none;
  box-shadow: none;
  padding: 6px;
}


/* Workaround Adwaita adding borders we don't want */
textview border.left {
  background: none;
}

treeview.dim-label {
  color: alpha(currentColor, 0.5);
}


button.run-arrow-button {
  min-width: 12px;
}


buildsurface list.sidebar row:selected button:hover {
  border-color: transparent;
  box-shadow: none;
  background: transparent;
  color: @theme_selected_fg_color;
  opacity: 1;
}
buildsurface list.sidebar row:selected button,
buildsurface list.sidebar row:selected button:active {
  opacity: 0.8;
}
buildsurface list.sidebar {
  border-right: 1px solid alpha(@borders, 0.55);
}


configurationview list row {
  padding: 10px;
  border-bottom: 1px solid alpha(@borders, 0.4);
}
configurationview list row:last-child {
  border-bottom: none;
}

configurationview list row entry {
  background: transparent;
  border: none;
  padding: 0;
  padding-left: 5px;
  border-radius: 3px;
  margin: -5px;
}

configurationview list row spinbutton entry {
  margin-left: 2px;
}

configurationview list row button {
  margin: -5px 0;
}


/* hrmm, we can use this to get row separators */
configurationview treeview {
  border-bottom: 1px solid alpha(@borders, 0.4);
}


popover.transfers list {
  background-color: transparent;
}
popover.transfers list row {
  border-top: 1px solid @borders;
}
popover.transfers list row:first-child {
  border-top: none;
}
popover.transfers list row > box {
  padding: 10px;
}

button.run-arrow-button {
  padding: 0px;
}


/* Stack list history tweaks */
dzlstacklist row {
  padding: 0;
  margin: 0;
}
dzlstacklist .stack-header {
  opacity: 0.55;
}

row:selected label.keycap {
  color: @theme_fg_color;
}
