/*
 * Greeter tweaks
 *
 * The following tweaks the greeter surface by adding
 * separator lines to the list box.
 */
.greeter list row {
  border-bottom: 1px solid alpha(@borders, 0.2);
  transition-property: background, opacity;
}
.greeter list row:last-child {
  border-bottom: none;
}
.greeter frame border {
  border-color: alpha(@borders, 0.6);
}

