treeview.i-wanna-be-listbox {
  color: @theme_fg_color;
  background: transparent;
}
treeview.i-wanna-be-listbox:hover {
  background-color: alpha(@theme_fg_color, 0.05);
}
treeview.i-wanna-be-listbox:selected {
  color: @theme_selected_fg_color;
  background-color: @theme_selected_bg_color;
}
treeview.i-wanna-be-listbox:backdrop {
  color: @theme_unfocused_fg_color;
  background-color: @theme_unfocused_base_color;
}
treeview.i-wanna-be-listbox:selected:backdrop {
  color: @theme_unfocused_selected_fg_color;
  background-color: @theme_unfocused_selected_bg_color;
}
